///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoOrderSpace.cc
// ------------------
// Cego order space implementation
//     
// Design and Implementation by Bjoern Lemke
//     
// (C)opyright 2000-2010 Bjoern Lemke
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// IMPLEMENTATION MODULE
//
// Class: CegoOrderSpace
// 
// Description: 
//
// Status: QG-2.6
//
///////////////////////////////////////////////////////////////////////////////

// base includes
#include <lfc/Exception.h>

// cego includes
#include "CegoOrderSpace.h"
#include "CegoDataType.h"
#include "CegoTableManager.h"

#include <string.h>
#include <stdlib.h>

CegoOrderSpace::CegoOrderSpace()
{
    _rootNode = 0;
}

CegoOrderSpace::~CegoOrderSpace()
{  
}

void CegoOrderSpace::createOrderSpace(ListT<CegoExpr*>* pOrderList, ListT<Ordering>* pOrderOptList, long maxOrderSize)
{

    dropOrderSpace();

    _pOrderList = pOrderList;
    _pOrderOptList = pOrderOptList;
    _maxOrderSize = maxOrderSize;
    _orderSize = 0;

    CegoExpr **pExpr = pOrderList->First();
    int id=1;
    while ( pExpr )
    {
	ListT<CegoAggregation*> aggList = (*pExpr)->getAggregationList();
	if ( aggList.Size() == 0 )
	{
	    // if no aggregation, we take all used attributes in order schema

	    ListT<CegoAttrDesc*> attrRefList = (*pExpr)->getAttrRefList();
	    
	    CegoAttrDesc** pAttrRef = attrRefList.First();
	    
	    while ( pAttrRef )
	    {
		CegoField f((*pAttrRef)->getTableName(), (*pAttrRef)->getAttrName());
		f.setId(id);
		_orderSchema.Insert(f);
		id++;
		pAttrRef = attrRefList.Next();
	    }
	}
	else
	{
	    // if aggregation, this is a grouping order, we just take the aggregated values

	    CegoAggregation **pAgg = aggList.First();
	    while ( pAgg )
	    {	    
		CegoField f;
		f.setAttrName("AGG");
		f.setId((*pAgg)->getAggregationId());
		_orderSchema.Insert(f);
		
		pAgg = aggList.Next();
	    }
	}
	pExpr = pOrderList->Next();
    }
}

void CegoOrderSpace::dropOrderSpace()
{    
    if ( _rootNode )
    {
	delete _rootNode;
	_rootNode = 0;
	_orderSize = 0;
    }
}

void CegoOrderSpace::insertTuple(ListT<CegoField>& addKey, ListT<CegoField>& addTuple)
{

    CegoField* pKF = addKey.First();
    CegoField* pOF = _orderSchema.First();
    while (pKF && pOF )
    {        
        pKF->setId(pOF->getId());
        pKF = addKey.Next();
        pOF = _orderSchema.Next();        
    }

    int id=1;
    CegoField *pF = addTuple.First();
    while (pF)
    {
	pF->setId(id);
	id++;
	pF = addTuple.Next();
    }

    if ( _rootNode == 0 )
    {

	ListT<CegoFieldValue> orderKey;
	
	CegoExpr **pExpr = _pOrderList->First();
	while ( pExpr )
	{
	    setAggregationValue(*pExpr, addKey);	
	    (*pExpr)->setFieldListArray(&addKey, 1);

	    (*pExpr)->clearAttrCache();

	    CegoFieldValue fv = (*pExpr)->evalFieldValue();
	    
	    orderKey.Insert(fv);
	    
	    pExpr = _pOrderList->Next();
	}
	
	_rootNode = new CegoOrderNode(orderKey, addTuple);      
	_orderSize = sizeof(*_rootNode);
    }
    else
    {
	insertTupleInTree(addKey, addTuple);
    }
}

CegoOrderCursor* CegoOrderSpace::getCursor()
{
    if ( _rootNode == 0 )
	throw Exception(EXLOC, "Orderspace not created, get cursor");

    CegoOrderCursor *pOC = new CegoOrderCursor(_rootNode);
    return pOC;
}

void CegoOrderSpace::insertTupleInTree(ListT<CegoField>& addKey, ListT<CegoField>& addTuple)
{

    ListT<CegoFieldValue> orderKey;

    CegoExpr **pExpr = _pOrderList->First();
    while ( pExpr )
    {
	setAggregationValue(*pExpr, addKey);	
	(*pExpr)->setFieldListArray(&addKey, 1);

	(*pExpr)->clearAttrCache();

	CegoFieldValue fv = (*pExpr)->evalFieldValue();
	
	orderKey.Insert(fv);

	pExpr = _pOrderList->Next();
    }

    CegoOrderNode* pNode = new CegoOrderNode(orderKey, addTuple);
    _orderSize += sizeof(*pNode);

    if ( _orderSize > _maxOrderSize )
    {
	delete pNode;
	throw Exception(EXLOC, "Order buffer exceeded");
    }

    CegoOrderNode* actNode = _rootNode;
            
    while ( 1 ) 
    {
	if  ( compKey(actNode->getKey(), orderKey) == HIGHER ) 
	{
	    if (  actNode->Left() == 0 )
	    {
		pNode->setParent(actNode);
		actNode->setLeft(pNode);		
		return;
	    }
	    else
	    {
		actNode = actNode->Left();
	    }
	}
	else
	{
	    if ( actNode->Right() == 0 )
	    {
		// insert at right
		pNode->setParent(actNode);
		actNode->setRight(pNode);		
		return;		
	    }
	    else
	    {
		actNode = actNode->Right();
	    }	    
	}	
    }
}

CegoOrderSpace::KeyComp CegoOrderSpace::compKey(ListT<CegoFieldValue>& key1, ListT<CegoFieldValue>& key2)
{

    Ordering *pOrd = _pOrderOptList->First();
    CegoFieldValue* pFV1 = key1.First();
    CegoFieldValue* pFV2 = key2.First();

    while ( pOrd && pFV1 && pFV2 )
    {
	if ( *pFV1 < *pFV2 )
	{
	    if ( *pOrd == ASC )
		return LOWER;
	    else
		return HIGHER;
	}
	else if ( *pFV1 > *pFV2 )
	{
	    if ( *pOrd == ASC )
		return HIGHER;
	    else
		return LOWER;
	}

	pOrd = _pOrderOptList->Next();
	pFV1 = key1.Next();
	pFV2 = key2.Next();
    }

    return EQUAL;
}

void CegoOrderSpace::setAggregationValue(CegoExpr *pExpr, ListT<CegoField>& fl)
{
    ListT<CegoAggregation*> aggList = pExpr->getAggregationList();
    
    CegoAggregation **pAgg = aggList.First();
    while ( pAgg )
    {
	CegoField *pF = fl.First();
	while ( pF )
	{
	    if ( pF->getId() == (*pAgg)->getAggregationId() )
	    {	
		(*pAgg)->setFieldValue(pF->getValue());
		pF = 0;
	    }
	    else
	    {
		pF = fl.Next();
	    }
	}
	pAgg = aggList.Next();
    }
}
